@@ -14,7 +14,7 @@ namespace duckdb {
 
 class JsonExtension : public Extension {
 public:
-	void Load(DuckDB &db) override;
+	void Load(ExtensionLoader &db) override;
 	std::string Name() override;
 	std::string Version() const override;
 };
