@@ -14,6 +14,7 @@ from .require_env import RequireEnv
 from .restart import Restart
 from .reconnect import Reconnect
 from .sleep import Sleep, SleepUnit
+from .unzip import Unzip
 
 from .skip import Skip, Unskip
 
@@ -35,6 +36,7 @@ __all__ = [
     Sleep,
     SleepUnit,
     Skip,
+    Unzip,
     Unskip,
     SortStyle,
 ]
