@@ -19,36 +19,44 @@ using ValidityBytes = TupleDataLayout::ValidityBytes;
 GroupedAggregateHashTable::GroupedAggregateHashTable(ClientContext &context, Allocator &allocator,
                                                      vector<LogicalType> group_types, vector<LogicalType> payload_types,
                                                      const vector<BoundAggregateExpression *> &bindings,
-                                                     idx_t initial_capacity, idx_t radix_bits)
+                                                     idx_t initial_capacity, idx_t radix_bits,
+                                                     TupleDataValidityType group_validity)
     : GroupedAggregateHashTable(context, allocator, std::move(group_types), std::move(payload_types),
-                                AggregateObject::CreateAggregateObjects(bindings), initial_capacity, radix_bits) {
+                                AggregateObject::CreateAggregateObjects(bindings), initial_capacity, radix_bits,
+                                group_validity) {
 }
 
 GroupedAggregateHashTable::GroupedAggregateHashTable(ClientContext &context, Allocator &allocator,
-                                                     vector<LogicalType> group_types)
-    : GroupedAggregateHashTable(context, allocator, std::move(group_types), {}, vector<AggregateObject>()) {
+                                                     vector<LogicalType> group_types,
+                                                     TupleDataValidityType group_validity)
+    : GroupedAggregateHashTable(context, allocator, std::move(group_types), {}, vector<AggregateObject>(),
+                                InitialCapacity(), 0, group_validity) {
 }
 
-GroupedAggregateHashTable::AggregateHTAppendState::AggregateHTAppendState()
-    : ht_offsets(LogicalType::UBIGINT), hash_salts(LogicalType::HASH), group_compare_vector(STANDARD_VECTOR_SIZE),
-      no_match_vector(STANDARD_VECTOR_SIZE), empty_vector(STANDARD_VECTOR_SIZE), new_groups(STANDARD_VECTOR_SIZE),
-      addresses(LogicalType::POINTER) {
+GroupedAggregateHashTable::AggregateHTAppendState::AggregateHTAppendState(ArenaAllocator &allocator)
+    : hashes(LogicalType::HASH), ht_offsets(LogicalType::UBIGINT), hash_salts(LogicalType::HASH),
+      new_groups(STANDARD_VECTOR_SIZE), group_compare_vector(STANDARD_VECTOR_SIZE),
+      no_match_vector(STANDARD_VECTOR_SIZE), addresses(LogicalType::POINTER), row_state(allocator) {
 }
 
-GroupedAggregateHashTable::GroupedAggregateHashTable(ClientContext &context, Allocator &allocator,
+GroupedAggregateHashTable::GroupedAggregateHashTable(ClientContext &context_p, Allocator &allocator,
                                                      vector<LogicalType> group_types_p,
                                                      vector<LogicalType> payload_types_p,
                                                      vector<AggregateObject> aggregate_objects_p,
-                                                     idx_t initial_capacity, idx_t radix_bits)
-    : BaseAggregateHashTable(context, allocator, aggregate_objects_p, std::move(payload_types_p)),
-      radix_bits(radix_bits), count(0), capacity(0), skip_lookups(false),
-      aggregate_allocator(make_shared_ptr<ArenaAllocator>(allocator)) {
+                                                     idx_t initial_capacity, idx_t radix_bits,
+                                                     TupleDataValidityType group_validity)
+    : BaseAggregateHashTable(context_p, allocator, aggregate_objects_p, std::move(payload_types_p)), context(context_p),
+      radix_bits(radix_bits), count(0), capacity(0), sink_count(0), skip_lookups(false), enable_hll(false),
+      aggregate_allocator(make_shared_ptr<ArenaAllocator>(allocator)), state(*aggregate_allocator) {
 
 	// Append hash column to the end and initialise the row layout
 	group_types_p.emplace_back(LogicalType::HASH);
-	layout.Initialize(std::move(group_types_p), std::move(aggregate_objects_p));
 
-	hash_offset = layout.GetOffsets()[layout.ColumnCount() - 1];
+	auto layout = make_shared_ptr<TupleDataLayout>();
+	layout->Initialize(std::move(group_types_p), std::move(aggregate_objects_p), group_validity);
+	layout_ptr = std::move(layout);
+
+	hash_offset = layout_ptr->GetOffsets()[layout_ptr->ColumnCount() - 1];
 
 	// Partitioned data and pointer table
 	InitializePartitionedData();
@@ -58,8 +66,10 @@ GroupedAggregateHashTable::GroupedAggregateHashTable(ClientContext &context, All
 	Resize(initial_capacity);
 
 	// Predicates
-	predicates.resize(layout.ColumnCount() - 1, ExpressionType::COMPARE_NOT_DISTINCT_FROM);
-	row_matcher.Initialize(true, layout, predicates);
+	const auto expr_type =
+	    layout_ptr->AllValid() ? ExpressionType::COMPARE_EQUAL : ExpressionType::COMPARE_NOT_DISTINCT_FROM;
+	predicates.resize(layout_ptr->ColumnCount() - 1, expr_type);
+	row_matcher.Initialize(true, *layout_ptr, predicates);
 }
 
 void GroupedAggregateHashTable::InitializePartitionedData() {
@@ -67,14 +77,14 @@ void GroupedAggregateHashTable::InitializePartitionedData() {
 	    RadixPartitioning::RadixBitsOfPowerOfTwo(partitioned_data->PartitionCount()) != radix_bits) {
 		D_ASSERT(!partitioned_data || partitioned_data->Count() == 0);
 		partitioned_data =
-		    make_uniq<RadixPartitionedTupleData>(buffer_manager, layout, radix_bits, layout.ColumnCount() - 1);
+		    make_uniq<RadixPartitionedTupleData>(buffer_manager, layout_ptr, radix_bits, layout_ptr->ColumnCount() - 1);
 	} else {
 		partitioned_data->Reset();
 	}
 
-	D_ASSERT(GetLayout().GetAggrWidth() == layout.GetAggrWidth());
-	D_ASSERT(GetLayout().GetDataWidth() == layout.GetDataWidth());
-	D_ASSERT(GetLayout().GetRowWidth() == layout.GetRowWidth());
+	D_ASSERT(GetLayout().GetAggrWidth() == layout_ptr->GetAggrWidth());
+	D_ASSERT(GetLayout().GetDataWidth() == layout_ptr->GetDataWidth());
+	D_ASSERT(GetLayout().GetRowWidth() == layout_ptr->GetRowWidth());
 
 	partitioned_data->InitializeAppendState(state.partitioned_append_state,
 	                                        TupleDataPinProperties::KEEP_EVERYTHING_PINNED);
@@ -84,7 +94,7 @@ void GroupedAggregateHashTable::InitializeUnpartitionedData() {
 	D_ASSERT(radix_bits >= UNPARTITIONED_RADIX_BITS_THRESHOLD);
 	if (!unpartitioned_data) {
 		unpartitioned_data =
-		    make_uniq<RadixPartitionedTupleData>(buffer_manager, layout, 0ULL, layout.ColumnCount() - 1);
+		    make_uniq<RadixPartitionedTupleData>(buffer_manager, layout_ptr, 0ULL, layout_ptr->ColumnCount() - 1);
 	} else {
 		unpartitioned_data->Reset();
 	}
@@ -106,7 +116,7 @@ unique_ptr<PartitionedTupleData> GroupedAggregateHashTable::AcquirePartitionedDa
 		if (unpartitioned_data) {
 			unpartitioned_data->FlushAppendState(state.unpartitioned_append_state);
 			unpartitioned_data->Unpin();
-			unpartitioned_data->Repartition(*partitioned_data);
+			unpartitioned_data->Repartition(context, *partitioned_data);
 		}
 		InitializeUnpartitionedData();
 	}
@@ -123,7 +133,7 @@ void GroupedAggregateHashTable::Abandon() {
 		if (unpartitioned_data) {
 			unpartitioned_data->FlushAppendState(state.unpartitioned_append_state);
 			unpartitioned_data->Unpin();
-			unpartitioned_data->Repartition(*partitioned_data);
+			unpartitioned_data->Repartition(context, *partitioned_data);
 		}
 		InitializeUnpartitionedData();
 	}
@@ -139,7 +149,7 @@ void GroupedAggregateHashTable::Abandon() {
 void GroupedAggregateHashTable::Repartition() {
 	auto old = AcquirePartitionedData();
 	D_ASSERT(old->GetPartitions().size() != partitioned_data->GetPartitions().size());
-	old->Repartition(*partitioned_data);
+	old->Repartition(context, *partitioned_data);
 }
 
 shared_ptr<ArenaAllocator> GroupedAggregateHashTable::GetAggregateAllocator() {
@@ -151,14 +161,13 @@ GroupedAggregateHashTable::~GroupedAggregateHashTable() {
 }
 
 void GroupedAggregateHashTable::Destroy() {
-	if (!partitioned_data || partitioned_data->Count() == 0 || !layout.HasDestructor()) {
+	if (!partitioned_data || partitioned_data->Count() == 0 || !layout_ptr->HasDestructor()) {
 		return;
 	}
 
 	// There are aggregates with destructors: Call the destructor for each of the aggregates
 	// Currently does not happen because aggregate destructors are called while scanning in RadixPartitionedHashTable
 	// LCOV_EXCL_START
-	RowOperationsState row_state(*aggregate_allocator);
 	for (auto &data_collection : partitioned_data->GetPartitions()) {
 		if (data_collection->Count() == 0) {
 			continue;
@@ -166,13 +175,17 @@ void GroupedAggregateHashTable::Destroy() {
 		TupleDataChunkIterator iterator(*data_collection, TupleDataPinProperties::DESTROY_AFTER_DONE, false);
 		auto &row_locations = iterator.GetChunkState().row_locations;
 		do {
-			RowOperations::DestroyStates(row_state, layout, row_locations, iterator.GetCurrentChunkCount());
+			RowOperations::DestroyStates(state.row_state, *layout_ptr, row_locations, iterator.GetCurrentChunkCount());
 		} while (iterator.Next());
 		data_collection->Reset();
 	}
 	// LCOV_EXCL_STOP
 }
 
+shared_ptr<TupleDataLayout> GroupedAggregateHashTable::GetLayoutPtr() {
+	return partitioned_data->GetLayoutPtr();
+}
+
 const TupleDataLayout &GroupedAggregateHashTable::GetLayout() const {
 	return partitioned_data->GetLayout();
 }
@@ -241,16 +254,38 @@ idx_t GroupedAggregateHashTable::GetSinkCount() const {
 	return sink_count;
 }
 
+idx_t GroupedAggregateHashTable::GetMaterializedCount() const {
+	auto result = partitioned_data->Count();
+	if (unpartitioned_data) {
+		result += unpartitioned_data->Count();
+	}
+	return result;
+}
+
 void GroupedAggregateHashTable::SkipLookups() {
 	skip_lookups = true;
 }
 
+void GroupedAggregateHashTable::EnableHLL(bool enable) {
+	enable_hll = enable;
+}
+
+bool GroupedAggregateHashTable::HLLEnabled() const {
+	return enable_hll;
+}
+
+idx_t GroupedAggregateHashTable::GetHLLUpperBound() const {
+	D_ASSERT(enable_hll);
+	return LossyNumericCast<idx_t>((1 + HyperLogLog::GetErrorRate()) * static_cast<double>(hll.Count()));
+}
+
 void GroupedAggregateHashTable::Resize(idx_t size) {
 	D_ASSERT(size >= STANDARD_VECTOR_SIZE);
 	D_ASSERT(IsPowerOfTwo(size));
 	if (Count() != 0 && size < capacity) {
 		throw InternalException("Cannot downsize a non-empty hash table!");
 	}
+	D_ASSERT(Count() == 0 || Count() == GetMaterializedCount());
 
 	capacity = size;
 	hash_map = buffer_manager.GetBufferAllocator().Allocate(capacity * sizeof(ht_entry_t));
@@ -268,6 +303,14 @@ void GroupedAggregateHashTable::Resize(idx_t size) {
 	Verify();
 }
 
+static void SaltIncrementAndWrap(idx_t &offset, const idx_t &salt, const idx_t &capacity_mask) {
+	// How many of the uppermost bits of the salt to determine the linear probing increment
+	static constexpr idx_t INCREMENT_BITS = 5;
+	// Extract the bits and make sure it's odd so we cover the entire domain when adding modulo a power of two
+	offset += (salt >> (64 - INCREMENT_BITS)) | 1;
+	offset &= capacity_mask;
+}
+
 void GroupedAggregateHashTable::ReinsertTuples(PartitionedTupleData &data) {
 	for (auto &data_collection : data.GetPartitions()) {
 		if (data_collection->Count() == 0) {
@@ -279,16 +322,17 @@ void GroupedAggregateHashTable::ReinsertTuples(PartitionedTupleData &data) {
 			for (idx_t i = 0; i < iterator.GetCurrentChunkCount(); i++) {
 				const auto &row_location = row_locations[i];
 				const auto hash = Load<hash_t>(row_location + hash_offset);
+				const auto salt = ht_entry_t::ExtractSalt(hash);
 
 				// Find an empty entry
 				auto ht_offset = ApplyBitMask(hash);
 				D_ASSERT(ht_offset == hash % capacity);
 				while (entries[ht_offset].IsOccupied()) {
-					IncrementAndWrap(ht_offset, bitmask);
+					SaltIncrementAndWrap(ht_offset, salt, bitmask);
 				}
 				auto &entry = entries[ht_offset];
 				D_ASSERT(!entry.IsOccupied());
-				entry.SetSalt(ht_entry_t::ExtractSalt(hash));
+				entry.SetSalt(salt);
 				entry.SetPointer(row_location);
 				D_ASSERT(entry.IsOccupied());
 			}
@@ -299,7 +343,7 @@ void GroupedAggregateHashTable::ReinsertTuples(PartitionedTupleData &data) {
 idx_t GroupedAggregateHashTable::AddChunk(DataChunk &groups, DataChunk &payload, AggregateType filter) {
 	unsafe_vector<idx_t> aggregate_filter;
 
-	auto &aggregates = layout.GetAggregates();
+	auto &aggregates = layout_ptr->GetAggregates();
 	for (idx_t i = 0; i < aggregates.size(); i++) {
 		auto &aggregate = aggregates[i];
 		if (aggregate.aggr_type == filter) {
@@ -329,7 +373,7 @@ optional_idx GroupedAggregateHashTable::TryAddDictionaryGroups(DataChunk &groups
 	if (dictionary_id.empty()) {
 		// dictionary has no id, we can't cache across vectors
 		// only use dictionary compression if there are fewer entries than groups
-		if (dict_size >= groups.size() * DICTIONARY_THRESHOLD) {
+		if (dict_size * DICTIONARY_THRESHOLD >= groups.size()) {
 			// dictionary is too large - use regular aggregation
 			return optional_idx();
 		}
@@ -387,7 +431,7 @@ optional_idx GroupedAggregateHashTable::TryAddDictionaryGroups(DataChunk &groups
 		// add the dictionary groups to the hash table
 		new_group_count = FindOrCreateGroups(unique_values, hashes, new_dictionary_pointers, state.new_groups);
 	}
-	auto &aggregates = layout.GetAggregates();
+	auto &aggregates = layout_ptr->GetAggregates();
 	if (aggregates.empty()) {
 		// early-out - no aggregates to update
 		return new_group_count;
@@ -400,7 +444,7 @@ optional_idx GroupedAggregateHashTable::TryAddDictionaryGroups(DataChunk &groups
 	auto dict_addresses = FlatVector::GetData<uintptr_t>(dictionary_addresses);
 	for (idx_t i = 0; i < unique_count; i++) {
 		auto dict_idx = unique_entries.get_index(i);
-		dict_addresses[dict_idx] = new_dict_addresses[i] + layout.GetAggrOffset();
+		dict_addresses[dict_idx] = new_dict_addresses[i] + layout_ptr->GetAggrOffset();
 	}
 	// now set up the addresses for the aggregates
 	auto result_addresses = FlatVector::GetData<uintptr_t>(state.addresses);
@@ -440,7 +484,7 @@ optional_idx GroupedAggregateHashTable::TryAddConstantGroups(DataChunk &groups,
 	auto &new_dictionary_pointers = dict_state.new_dictionary_pointers;
 	auto new_group_count = FindOrCreateGroups(unique_values, hashes, new_dictionary_pointers, state.new_groups);
 
-	auto &aggregates = layout.GetAggregates();
+	auto &aggregates = layout_ptr->GetAggregates();
 	if (aggregates.empty()) {
 		// early-out - no aggregates to update
 		return new_group_count;
@@ -448,7 +492,7 @@ optional_idx GroupedAggregateHashTable::TryAddConstantGroups(DataChunk &groups,
 
 	auto new_dict_addresses = FlatVector::GetData<uintptr_t>(new_dictionary_pointers);
 	auto result_addresses = FlatVector::GetData<uintptr_t>(state.addresses);
-	uintptr_t aggregate_address = new_dict_addresses[0] + layout.GetAggrOffset();
+	uintptr_t aggregate_address = new_dict_addresses[0] + layout_ptr->GetAggrOffset();
 	for (idx_t i = 0; i < payload.size(); i++) {
 		result_addresses[i] = aggregate_address;
 	}
@@ -466,7 +510,8 @@ optional_idx GroupedAggregateHashTable::TryAddCompressedGroups(DataChunk &groups
 	if (groups.AllConstant()) {
 		return TryAddConstantGroups(groups, payload, filter);
 	}
-	if (groups.ColumnCount() == 1 && groups.data[0].GetVectorType() == VectorType::DICTIONARY_VECTOR) {
+	if (groups.ColumnCount() == 1 && groups.data[0].GetVectorType() == VectorType::DICTIONARY_VECTOR &&
+	    groups.data[0].GetType().InternalType() != PhysicalType::STRUCT) {
 		return TryAddDictionaryGroups(groups, payload, filter);
 	}
 	return optional_idx();
@@ -481,18 +526,16 @@ idx_t GroupedAggregateHashTable::AddChunk(DataChunk &groups, DataChunk &payload,
 		return result.GetIndex();
 	}
 	// otherwise append the raw values
-	Vector hashes(LogicalType::HASH);
-	groups.Hash(hashes);
+	groups.Hash(state.hashes);
 
-	return AddChunk(groups, hashes, payload, filter);
+	return AddChunk(groups, state.hashes, payload, filter);
 }
 
 void GroupedAggregateHashTable::UpdateAggregates(DataChunk &payload, const unsafe_vector<idx_t> &filter) {
 	// Now every cell has an entry, update the aggregates
-	auto &aggregates = layout.GetAggregates();
+	auto &aggregates = layout_ptr->GetAggregates();
 	idx_t filter_idx = 0;
 	idx_t payload_idx = 0;
-	RowOperationsState row_state(*aggregate_allocator);
 	for (idx_t i = 0; i < aggregates.size(); i++) {
 		auto &aggr = aggregates[i];
 		if (filter_idx >= filter.size() || i < filter[filter_idx]) {
@@ -504,10 +547,10 @@ void GroupedAggregateHashTable::UpdateAggregates(DataChunk &payload, const unsaf
 		D_ASSERT(i == filter[filter_idx]);
 
 		if (aggr.aggr_type != AggregateType::DISTINCT && aggr.filter) {
-			RowOperations::UpdateFilteredStates(row_state, filter_set.GetFilterData(i), aggr, state.addresses, payload,
-			                                    payload_idx);
+			RowOperations::UpdateFilteredStates(state.row_state, filter_set.GetFilterData(i), aggr, state.addresses,
+			                                    payload, payload_idx);
 		} else {
-			RowOperations::UpdateStates(row_state, aggr, state.addresses, payload, payload_idx, payload.size());
+			RowOperations::UpdateStates(state.row_state, aggr, state.addresses, payload, payload_idx, payload.size());
 		}
 
 		// Move to the next aggregate
@@ -526,14 +569,14 @@ idx_t GroupedAggregateHashTable::AddChunk(DataChunk &groups, Vector &group_hashe
 	}
 
 #ifdef DEBUG
-	D_ASSERT(groups.ColumnCount() + 1 == layout.ColumnCount());
+	D_ASSERT(groups.ColumnCount() + 1 == layout_ptr->ColumnCount());
 	for (idx_t i = 0; i < groups.ColumnCount(); i++) {
-		D_ASSERT(groups.GetTypes()[i] == layout.GetTypes()[i]);
+		D_ASSERT(groups.GetTypes()[i] == layout_ptr->GetTypes()[i]);
 	}
 #endif
 
 	const auto new_group_count = FindOrCreateGroups(groups, group_hashes, state.addresses, state.new_groups);
-	VectorOperations::AddInPlace(state.addresses, NumericCast<int64_t>(layout.GetAggrOffset()), payload.size());
+	VectorOperations::AddInPlace(state.addresses, NumericCast<int64_t>(layout_ptr->GetAggrOffset()), payload.size());
 
 	UpdateAggregates(payload, filter);
 
@@ -543,7 +586,7 @@ idx_t GroupedAggregateHashTable::AddChunk(DataChunk &groups, Vector &group_hashe
 void GroupedAggregateHashTable::FetchAggregates(DataChunk &groups, DataChunk &result) {
 #ifdef DEBUG
 	groups.Verify();
-	D_ASSERT(groups.ColumnCount() + 1 == layout.ColumnCount());
+	D_ASSERT(groups.ColumnCount() + 1 == layout_ptr->ColumnCount());
 	for (idx_t i = 0; i < result.ColumnCount(); i++) {
 		D_ASSERT(result.data[i].GetType() == payload_types[i]);
 	}
@@ -556,16 +599,49 @@ void GroupedAggregateHashTable::FetchAggregates(DataChunk &groups, DataChunk &re
 
 	// find the groups associated with the addresses
 	// FIXME: this should not use the FindOrCreateGroups, creating them is unnecessary
-	Vector addresses(LogicalType::POINTER);
-	FindOrCreateGroups(groups, addresses);
+	FindOrCreateGroups(groups, state.addresses);
 	// now fetch the aggregates
-	RowOperationsState row_state(*aggregate_allocator);
-	RowOperations::FinalizeStates(row_state, layout, addresses, result, 0);
+	RowOperations::FinalizeStates(state.row_state, *layout_ptr, state.addresses, result, 0);
+}
+
+template <bool HAS_SEL>
+static void GroupedAggregateHashTableInnerLoop(ht_entry_t *const entries, const idx_t capacity, const hash_t bitmask,
+                                               const hash_t *const hash_salts, uint64_t *const ht_offsets,
+                                               const SelectionVector *const sel_vector, const idx_t remaining_entries,
+                                               SelectionVector &empty_vector, SelectionVector &compare_vector,
+                                               idx_t &empty_count, idx_t &compare_count) {
+	// For each remaining entry, figure out whether or not it belongs to a full or empty group
+	for (idx_t i = 0; i < remaining_entries; i++) {
+		const auto index = HAS_SEL ? sel_vector->get_index_unsafe(i) : i;
+		const auto salt = hash_salts[index];
+		auto &ht_offset = ht_offsets[index];
+
+		idx_t inner_iteration_count;
+		for (inner_iteration_count = 0; inner_iteration_count < capacity; inner_iteration_count++) {
+			auto &entry = entries[ht_offset];
+			if (!entry.IsOccupied()) { // Unoccupied: claim it
+				entry.SetSalt(salt);
+				empty_vector.set_index(empty_count++, index);
+				break;
+			}
+
+			if (DUCKDB_LIKELY(entry.GetSalt() == salt)) { // Matching salt: compare groups
+				compare_vector.set_index(compare_count++, index);
+				break;
+			}
+
+			// Linear probing
+			SaltIncrementAndWrap(ht_offset, salt, bitmask);
+		}
+		if (DUCKDB_UNLIKELY(inner_iteration_count == capacity)) {
+			throw InternalException("Maximum inner iteration count reached in GroupedAggregateHashTable");
+		}
+	}
 }
 
 idx_t GroupedAggregateHashTable::FindOrCreateGroupsInternal(DataChunk &groups, Vector &group_hashes_v,
                                                             Vector &addresses_v, SelectionVector &new_groups_out) {
-	D_ASSERT(groups.ColumnCount() + 1 == layout.ColumnCount());
+	D_ASSERT(groups.ColumnCount() + 1 == layout_ptr->ColumnCount());
 	D_ASSERT(group_hashes_v.GetType() == LogicalType::HASH);
 	D_ASSERT(state.ht_offsets.GetVectorType() == VectorType::FLAT_VECTOR);
 	D_ASSERT(state.ht_offsets.GetType() == LogicalType::UBIGINT);
@@ -585,9 +661,9 @@ idx_t GroupedAggregateHashTable::FindOrCreateGroupsInternal(DataChunk &groups, V
 
 	// Make a chunk that references the groups and the hashes and convert to unified format
 	if (state.group_chunk.ColumnCount() == 0) {
-		state.group_chunk.InitializeEmpty(layout.GetTypes());
+		state.group_chunk.InitializeEmpty(layout_ptr->GetTypes());
 	}
-	D_ASSERT(state.group_chunk.ColumnCount() == layout.GetTypes().size());
+	D_ASSERT(state.group_chunk.ColumnCount() == layout_ptr->GetTypes().size());
 	for (idx_t grp_idx = 0; grp_idx < groups.ColumnCount(); grp_idx++) {
 		state.group_chunk.data[grp_idx].Reference(groups.data[grp_idx]);
 	}
@@ -596,10 +672,10 @@ idx_t GroupedAggregateHashTable::FindOrCreateGroupsInternal(DataChunk &groups, V
 
 	// convert all vectors to unified format
 	TupleDataCollection::ToUnifiedFormat(state.partitioned_append_state.chunk_state, state.group_chunk);
-	if (!state.group_data) {
-		state.group_data = make_unsafe_uniq_array_uninitialized<UnifiedVectorFormat>(state.group_chunk.ColumnCount());
+
+	if (enable_hll) {
+		hll.Update(group_hashes_v, group_hashes_v, groups.size());
 	}
-	TupleDataCollection::GetVectorData(state.partitioned_append_state.chunk_state, state.group_data.get());
 
 	group_hashes_v.Flatten(chunk_size);
 	const auto hashes = FlatVector::GetData<hash_t>(group_hashes_v);
@@ -611,14 +687,14 @@ idx_t GroupedAggregateHashTable::FindOrCreateGroupsInternal(DataChunk &groups, V
 		// Just appending now
 		partitioned_data->AppendUnified(state.partitioned_append_state, state.group_chunk,
 		                                *FlatVector::IncrementalSelectionVector(), chunk_size);
-		RowOperations::InitializeStates(layout, state.partitioned_append_state.chunk_state.row_locations,
+		RowOperations::InitializeStates(*layout_ptr, state.partitioned_append_state.chunk_state.row_locations,
 		                                *FlatVector::IncrementalSelectionVector(), chunk_size);
 
 		const auto row_locations =
 		    FlatVector::GetData<data_ptr_t>(state.partitioned_append_state.chunk_state.row_locations);
 		const auto &row_sel = state.partitioned_append_state.reverse_partition_sel;
 		for (idx_t i = 0; i < chunk_size; i++) {
-			const auto &row_idx = row_sel[i];
+			const auto row_idx = row_sel.get_index_unsafe(i);
 			const auto &row_location = row_locations[row_idx];
 			addresses[i] = row_location;
 		}
@@ -644,6 +720,7 @@ idx_t GroupedAggregateHashTable::FindOrCreateGroupsInternal(DataChunk &groups, V
 		hash_salts[r] = ht_entry_t::ExtractSalt(hash);
 	}
 
+	SelectionVector empty_vector;
 	idx_t new_group_count = 0;
 	idx_t remaining_entries = chunk_size;
 	idx_t iteration_count;
@@ -652,34 +729,19 @@ idx_t GroupedAggregateHashTable::FindOrCreateGroupsInternal(DataChunk &groups, V
 		idx_t need_compare_count = 0;
 		idx_t no_match_count = 0;
 
-		// For each remaining entry, figure out whether or not it belongs to a full or empty group
-		for (idx_t i = 0; i < remaining_entries; i++) {
-			const auto index = sel_vector->get_index(i);
-			const auto salt = hash_salts[index];
-			auto &ht_offset = ht_offsets[index];
-
-			idx_t inner_iteration_count;
-			for (inner_iteration_count = 0; inner_iteration_count < capacity; inner_iteration_count++) {
-				auto &entry = entries[ht_offset];
-				if (!entry.IsOccupied()) { // Unoccupied: claim it
-					entry.SetSalt(salt);
-					state.empty_vector.set_index(new_entry_count++, index);
-					new_groups_out.set_index(new_group_count++, index);
-					break;
-				}
-
-				if (DUCKDB_LIKELY(entry.GetSalt() == salt)) { // Matching salt: compare groups
-					state.group_compare_vector.set_index(need_compare_count++, index);
-					break;
-				}
-
-				// Linear probing
-				IncrementAndWrap(ht_offset, bitmask);
-			}
-			if (DUCKDB_UNLIKELY(inner_iteration_count == capacity)) {
-				throw InternalException("Maximum inner iteration count reached in GroupedAggregateHashTable");
-			}
+		// "new_groups_out" contains ALL groups for the chunk, "empty_vector" only the groups for this iteration,
+		// so it's just the same selection vector, but offset by the current "new_group_count".
+		empty_vector.Initialize(new_groups_out.data() + new_group_count);
+		if (sel_vector->IsSet()) {
+			GroupedAggregateHashTableInnerLoop<true>(entries, capacity, bitmask, hash_salts, ht_offsets, sel_vector,
+			                                         remaining_entries, empty_vector, state.group_compare_vector,
+			                                         new_entry_count, need_compare_count);
+		} else {
+			GroupedAggregateHashTableInnerLoop<false>(entries, capacity, bitmask, hash_salts, ht_offsets, sel_vector,
+			                                          remaining_entries, empty_vector, state.group_compare_vector,
+			                                          new_entry_count, need_compare_count);
 		}
+		new_group_count += new_entry_count;
 
 		if (DUCKDB_UNLIKELY(occupied_count > new_entry_count + need_compare_count)) {
 			// We use the useless occupied_count we summed above here so the variable is used,
@@ -701,16 +763,16 @@ idx_t GroupedAggregateHashTable::FindOrCreateGroupsInternal(DataChunk &groups, V
 				data = partitioned_data.get();
 				append_state = &state.partitioned_append_state;
 			}
-			data->AppendUnified(*append_state, state.group_chunk, state.empty_vector, new_entry_count);
-			RowOperations::InitializeStates(layout, append_state->chunk_state.row_locations,
+			data->AppendUnified(*append_state, state.group_chunk, empty_vector, new_entry_count);
+			RowOperations::InitializeStates(*layout_ptr, append_state->chunk_state.row_locations,
 			                                *FlatVector::IncrementalSelectionVector(), new_entry_count);
 
 			// Set the entry pointers in the 1st part of the HT now that the data has been appended
 			const auto row_locations = FlatVector::GetData<data_ptr_t>(append_state->chunk_state.row_locations);
 			const auto &row_sel = append_state->reverse_partition_sel;
 			for (idx_t new_entry_idx = 0; new_entry_idx < new_entry_count; new_entry_idx++) {
-				const auto &index = state.empty_vector[new_entry_idx];
-				const auto &row_idx = row_sel[index];
+				const auto index = empty_vector.get_index_unsafe(new_entry_idx);
+				const auto row_idx = row_sel.get_index_unsafe(index);
 				const auto &row_location = row_locations[row_idx];
 
 				auto &entry = entries[ht_offsets[index]];
@@ -723,22 +785,23 @@ idx_t GroupedAggregateHashTable::FindOrCreateGroupsInternal(DataChunk &groups, V
 		if (need_compare_count != 0) {
 			// Get the pointers to the rows that need to be compared
 			for (idx_t need_compare_idx = 0; need_compare_idx < need_compare_count; need_compare_idx++) {
-				const auto &index = state.group_compare_vector[need_compare_idx];
+				const auto index = state.group_compare_vector.get_index_unsafe(need_compare_idx);
 				const auto &entry = entries[ht_offsets[index]];
 				addresses[index] = entry.GetPointer();
 			}
 
 			// Perform group comparisons
 			row_matcher.Match(state.group_chunk, state.partitioned_append_state.chunk_state.vector_data,
-			                  state.group_compare_vector, need_compare_count, layout, addresses_v,
-			                  &state.no_match_vector, no_match_count);
+			                  state.group_compare_vector, need_compare_count, addresses_v, &state.no_match_vector,
+			                  no_match_count);
 		}
 
 		// Linear probing: each of the entries that do not match move to the next entry in the HT
 		for (idx_t i = 0; i < no_match_count; i++) {
-			const auto &index = state.no_match_vector[i];
+			const auto index = state.no_match_vector.get_index_unsafe(i);
 			auto &ht_offset = ht_offsets[index];
-			IncrementAndWrap(ht_offset, bitmask);
+			const auto salt = hash_salts[index];
+			SaltIncrementAndWrap(ht_offset, salt, bitmask);
 		}
 		sel_vector = &state.no_match_vector;
 		remaining_entries = no_match_count;
@@ -765,9 +828,8 @@ void GroupedAggregateHashTable::FindOrCreateGroups(DataChunk &groups, Vector &ad
 
 idx_t GroupedAggregateHashTable::FindOrCreateGroups(DataChunk &groups, Vector &addresses_out,
                                                     SelectionVector &new_groups_out) {
-	Vector hashes(LogicalType::HASH);
-	groups.Hash(hashes);
-	return FindOrCreateGroups(groups, hashes, addresses_out, new_groups_out);
+	groups.Hash(state.hashes);
+	return FindOrCreateGroups(groups, state.hashes, addresses_out, new_groups_out);
 }
 
 struct FlushMoveState {
@@ -820,26 +882,29 @@ void GroupedAggregateHashTable::Combine(GroupedAggregateHashTable &other) {
 }
 
 void GroupedAggregateHashTable::Combine(TupleDataCollection &other_data, optional_ptr<atomic<double>> progress) {
-	D_ASSERT(other_data.GetLayout().GetAggrWidth() == layout.GetAggrWidth());
-	D_ASSERT(other_data.GetLayout().GetDataWidth() == layout.GetDataWidth());
-	D_ASSERT(other_data.GetLayout().GetRowWidth() == layout.GetRowWidth());
+	D_ASSERT(other_data.GetLayout().GetAggrWidth() == layout_ptr->GetAggrWidth());
+	D_ASSERT(other_data.GetLayout().GetDataWidth() == layout_ptr->GetDataWidth());
+	D_ASSERT(other_data.GetLayout().GetRowWidth() == layout_ptr->GetRowWidth());
 
 	if (other_data.Count() == 0) {
 		return;
 	}
 
 	FlushMoveState fm_state(other_data);
-	RowOperationsState row_state(*aggregate_allocator);
 
 	idx_t chunk_idx = 0;
 	const auto chunk_count = other_data.ChunkCount();
 	while (fm_state.Scan()) {
+		// Check for interrupts with each chunk
+		if (context.interrupted.load(std::memory_order_relaxed)) {
+			throw InterruptException();
+		}
 		const auto input_chunk_size = fm_state.groups.size();
 		FindOrCreateGroups(fm_state.groups, fm_state.hashes, fm_state.group_addresses, fm_state.new_groups_sel);
-		RowOperations::CombineStates(row_state, layout, fm_state.scan_state.chunk_state.row_locations,
+		RowOperations::CombineStates(state.row_state, *layout_ptr, fm_state.scan_state.chunk_state.row_locations,
 		                             fm_state.group_addresses, input_chunk_size);
-		if (layout.HasDestructor()) {
-			RowOperations::DestroyStates(row_state, layout, fm_state.scan_state.chunk_state.row_locations,
+		if (layout_ptr->HasDestructor()) {
+			RowOperations::DestroyStates(state.row_state, *layout_ptr, fm_state.scan_state.chunk_state.row_locations,
 			                             input_chunk_size);
 		}
 
@@ -851,4 +916,38 @@ void GroupedAggregateHashTable::Combine(TupleDataCollection &other_data, optiona
 	Verify();
 }
 
+void GroupedAggregateHashTable::InitializeScan(AggregateHTScanState &scan_state) {
+	scan_state.partition_idx = 0;
+	vector<idx_t> group_indexes(layout_ptr->ColumnCount() - 1);
+	for (idx_t i = 0; i < group_indexes.size(); i++) {
+		group_indexes[i] = i;
+	}
+
+	auto &partition = partitioned_data->GetPartitions()[scan_state.partition_idx];
+	partition->InitializeScan(scan_state.scan_states, group_indexes);
+}
+
+bool GroupedAggregateHashTable::Scan(AggregateHTScanState &scan_state, DataChunk &distinct_rows,
+                                     DataChunk &payload_rows) {
+	if (scan_state.partition_idx >= partitioned_data->PartitionCount()) {
+		return false;
+	}
+
+	payload_rows.Reset();
+	distinct_rows.Reset();
+	auto &current_partition = partitioned_data->GetPartitions()[scan_state.partition_idx];
+
+	if (current_partition->Scan(scan_state.scan_states, distinct_rows)) {
+		FetchAggregates(distinct_rows, payload_rows);
+		return true;
+	} else {
+		if (++(scan_state.partition_idx) >= partitioned_data->PartitionCount()) {
+			return false;
+		} else {
+			auto &new_partition = partitioned_data->GetPartitions()[scan_state.partition_idx];
+			new_partition->InitializeScan(scan_state.scan_states);
+			return true;
+		}
+	}
+}
 } // namespace duckdb
